#!/usr/bin/env ruby -w
# ex:ts=10000
# vim:sw=4:sts=4:tw=0

require 'test/unit'
require 'relation'

class Relation__test < Test::Unit::TestCase
    def setup
        @klass = Class.new(Relation)
    end
    # Test a table with a couple of columns
    def test_some_columns
        @klass.tablename = "Article"
        @klass.column :ArticleId, "INTEGER"
        @klass.column :Description, "TEXT"
        expected = <<EOM
CREATE TABLE Article (
    ArticleId INTEGER,
    Description TEXT
);
EOM
        assert_equal(expected, @klass.sql)
    end
    # Test a table with no columns
    def test_nocolumn
        @klass.tablename = "Article"
        assert_equal("CREATE TABLE Article (\n\n);\n", @klass.sql)
    end
    # Test for wrong column specifications
    def test_column_errors
        [[ArgumentError, "NoType"],
         [RuntimeError,  "With&IllegalChars", "INTEGER"],
         [RuntimeError,  "WithIllegalType",   "TYPE&ERROR"],
        ].each { |args|
            assert_raises(args.shift) {
               @klass.column(*args)
            }
        }
    end
    # Test that we can set the tablename of a table
    def test_tablename
        @klass.tablename = "Article"
        assert_equal("CREATE TABLE Article (\n\n);\n", @klass.sql)
        @klass.tablename = "Preposition"
        assert_equal("CREATE TABLE Preposition (\n\n);\n", @klass.sql)
    end
    #
    # Test of select's SQL generation, including partial test of order_by
    #
    def test_select
        @klass.tablename = "Article"
        @klass.column :description, "TEXT"
        @klass.reference(@klass, :parent)
        selector3 = @klass.select(:article_id => 3)
        # FIXME Raising an exception on a relation that can only be empty (ie,
        # two incompatible comparisons) might be useful, but it is presently
        # not supported.
        #assert_raises(RuntimeError) {
        #    selector3.select(:article_id => 4)
        #}
        assert_raises(RuntimeError) {
            selector3.select(:unknown_column => 1)
        }
        assert_equal("SELECT * FROM Article WHERE article_id=3", selector3._select_sql)
        selector2 = @klass.select(:article_id => 2)
        assert_equal("SELECT * FROM Article WHERE article_id=2", selector2._select_sql)
        assert_equal("SELECT * FROM Article WHERE article_id=3 ORDER BY article_id",
            selector3.order_by(:article_id)._select_sql)
        assert_equal("SELECT * FROM Article WHERE description='SOMETHING'",
            @klass.select(:description => "SOMETHING")._select_sql)
        assert_equal("SELECT * FROM Article WHERE parent=article_id",
            @klass.select(:parent => @klass.article_id)._select_sql)
        # Should only be able to pass through Relation classes
        assert_raises(RuntimeError) {
            @klass.select(:parent => Class.new)
        }
    end
    #
    # Test SQL quoting of strings including escapes
    #
    def test_sqlquote
        assert_equal("'test'", @klass._sqlquote("test"))
        assert_equal("'\\''",  @klass._sqlquote("'"))
        assert_equal("'\\\\'", @klass._sqlquote("\\"))
    end
    #
    # Test that various forms of columns can get generated, and that
    # they can be manipulated as appropriate afterwards
    #
    def test_column
        @klass.column :foo, "INTEGER"
        @klass.tablename = "Bar"
        assert_nil(@klass.compare_column)
        assert_nothing_raised {
            @klass.foo == 1
        }
        assert_nil(@klass.compare_column)
        assert_raises(NoMethodError) {
            @klass.bar == 1
        }
        assert_nil(@klass.compare_column)
        assert_equal("SELECT * FROM Bar WHERE foo=2",
            (@klass.foo == 2)._select_sql)
        assert_nil(@klass.compare_column)
        assert_raises(TypeError) {
            @klass < 3
        }
        assert_equal("SELECT * FROM Bar WHERE foo<3",
            (@klass.foo < 3)._select_sql)
        assert_equal("SELECT * FROM Bar WHERE foo>4",
            (@klass.foo > 4)._select_sql)
        assert_equal("SELECT * FROM Bar WHERE foo>3 AND foo<4",
            ((@klass.foo > 3).foo < 4)._select_sql)
    end
    def test_has_column?
        # FIXME Not implemented
    end
    def test_pkey
        # FIXME Not implemented
    end
end

#
# Test relations in the context of a more complex setup.
# Defines a complete database, and tests on the stuff available there.
#
class Relation_set__test < Test::Unit::TestCase
    class Deleteable < Relation
    end
    class OrderMock < Relation
        column(:foo, "INTEGER")
    end
    class Foo < Relation
        column :foothing, "TEXT"
    end
    class Bar < Relation
        reference(Foo)
        column :tag, "VARCHAR(255)"
        reference(Bar, :parent)
    end
    def test_leftjoin
        assert_nothing_raised {
            Foo.leftjoin(Bar)
        }
        # FIXME More tests here
    end
    def test_delete!
        assert_equal("DELETE FROM Deleteable", Deleteable._sql_delete!)
        assert_equal("DELETE FROM Deleteable WHERE deleteable_id=123",
            Deleteable.select(:deleteable_id => 123)._sql_delete!)
    end
    def test_finish!
        # Test auto-generation of id columns
        # FIXME Test not implmented yet; implictly tested as part of the above tests,
        # but should have an explicit test including supression by pkey()
    end
    def test_order_by
        # Existing columns are allowed
        assert_nothing_raised {
            OrderMock.order_by(:ordermock_id)
            OrderMock.order_by(:foo)
        }
        # Unknown are NOT allowed
        assert_raises(RuntimeError) {
            OrderMock.order_by(:unknown_column)
        }
        # Duplicates are not allowed
        assert_raises(RuntimeError) {
           OrderMock.order_by(:foo).order_by(:foo)
        }
        assert_raises(RuntimeError) {
           OrderMock.order_by(:foo, :foo)
        }
        outsql = nil
        assert_nothing_raised {
            outsql = OrderMock.order_by(:ordermock_id).order_by(:foo)._order_by_sql
        }
        assert_equal(" ORDER BY ordermock_id,foo", outsql)
        assert_nothing_raised {
            outsql = OrderMock.order_by(:ordermock_id, :foo)._order_by_sql
        }
        assert_equal(" ORDER BY ordermock_id,foo", outsql)
        assert_nothing_raised {
            outsql = OrderMock.order_by_desc(:ordermock_id).order_by(:foo)._order_by_sql
        }
        assert_equal(" ORDER BY ordermock_id DESC,foo", outsql)
        assert_nothing_raised {
            outsql = OrderMock.order_by_desc(:ordermock_id, :foo)._order_by_sql
        }
        assert_equal(" ORDER BY ordermock_id DESC,foo DESC", outsql)
        # How to handle joins?
    end
    #
    # Test various forms of setting up references to other tables
    #
    def test_reference
        klass = Class.new(Relation)
        assert_nothing_raised {
            klass.reference(Foo)
        }
        assert(klass.has_column?(:foo_id), "No reference column generated: foo_id")
        assert_nothing_raised {
            klass.reference(Foo, :foo2)
        }
        assert(klass.has_column?(:foo2), "No reference column generated: foo2")
    end
    def test_=
        klass = Class.new(Relation)
        klass.column(:foo, "INTEGER")
        klass.column(:bar, "TEXT")
        klass.column(:baz, "TEXT")
        # FIXME SHOULD raise, as we're doing an assignment to the primary object!
        assert_nothing_raised {
            klass.foo = 1
        }
        assert_raises(NoMethodError) {
            klass.unknown_column = 1
        }
        # Verify that we allow good assignments
        assert_nothing_raised {
            selector = klass.select
            selector.foo = 1
            selector.bar = "SOMETHING"
        }
        # Verify that we disallow assignment across domains
        assert_raises(RuntimeError) {
            selector = klass.select
            selector.foo = "SOMETHING"
        }
        assert_raises(RuntimeError) {
            selector = klass.select
            selector.bar = 1
        }
        # Verify that we get an error if we try to cross domains with column references
        assert_raises(RuntimeError) {
            selector = klass.select
            selector.bar = selector.foo
        }
        # Verify OK inside domains
        assert_nothing_raised {
            selector = klass.select
            selector.bar = selector.baz
        }

        selector = klass.select
        # Verify that we do not allow objects of unknown classes
        assert_raises(RuntimeError, "Can assign a plain object to a field") {
            selector.foo = Object.new
        }
        # Verify that we do not allow unknown Classes
        assert_raises(RuntimeError, "Can assign Object (the class) to a field") {
            selector.foo = Object
        }
        # Verify that we do not allow cross-table updates
        assert_raises(RuntimeError) {
            selector.foo = Bar.tag
        }
    end
    #
    # Test setting up updates
    #
    def test_update
        ["SOMETHING", "ANYTHING"].each { |matcher|
            selector = Bar.tag == matcher
            selector.parent = selector.bar_id
            assert_equal("UPDATE Bar SET parent=bar_id WHERE tag='#{matcher}'",
                selector._update_sql)
        }
        selector = OrderMock.ordermock_id == 1
        selector.foo = 12345
        assert_equal("UPDATE OrderMock SET foo=12345 WHERE ordermock_id=1",
            selector._update_sql)
    end
    #
    # Test doing a projection
    #
    def test_project
        assert_raises(RuntimeError) {
            Bar.project(:unknown_column)
        }
        columns = Bar.columns
        projected = Bar.project(:bar_id, :foo_id).columns
        assert_not_equal(columns, projected)
        should_be_projected = {
            "foo_id" => columns["foo_id"],
            "bar_id" => columns["bar_id"]
        }
        assert_equal(should_be_projected, projected)
    end
    #
    # Test natural joins
    #
    def test_mul
        relation = nil
        assert_nothing_raised {
            relation = Bar * Bar
        }
        # FIXME This is intended for a Bar * Foo comparison, but that is
        # presently not allowed.
        #assert_equal(["bar_id", "foo_id", "foothing", "parent", "tag"],
        #    relation.columns.keys.sort)
        relation = (Bar.foo_id == 1) * (Bar.tag == "SOMETHING")
        assert_equal("SELECT * FROM Bar WHERE foo_id=1 AND tag='SOMETHING'",
            relation._select_sql)
        # FIXME DOES NOT WORK
        # Test optimization of AND
        relation = (Bar.foo_id == 1) * (Bar.foo_id == 1)
        #assert_equal("SELECT * FROM Bar WHERE foo_id=1",
        #    relation._select_sql)
    end
    #
    # Test unions
    #
    def test_plus
        relation = nil
        assert_nothing_raised {
            relation = (Bar.foo_id == 1) + (Bar.foo_id == 2)
        }
        assert_equal("SELECT * FROM Bar WHERE foo_id=1 OR foo_id=2",
            relation._select_sql)
        assert_nothing_raised {
            relation = (Bar.foo_id == 1) + (Bar.tag == "SOMETHING")
        }
        assert_equal("SELECT * FROM Bar WHERE foo_id=1 OR tag='SOMETHING'",
            relation._select_sql)
        assert_raises(RuntimeError) {
            Foo + Bar
        }
    end
end

class Compare__test < Test::Unit::TestCase
    Compare = Relation::Compare
    def setup
        @c1 = Relation::Compare::Equal.new("foo", 1)
        @c1a = Relation::Compare::Equal.new("foo", 1)
        @c2 = Relation::Compare::Equal.new("foo", 2)
        @c2a = Relation::Compare::Equal.new("foo", 2)
        assert_equal(@c1, @c1a)
        assert_not_equal(@c1, @c2)
        assert(@c1.eql?(@c1a))
        assert(!@c1.eql?(@c2))
        assert_equal(@c1.hash, @c1a.hash)
    end
    def andor_paramtest(klass)
        assert_raises(RuntimeError) {
            klass.new(nil, @c2)
        }
        assert_raises(RuntimeError) {
            klass.new(@c1, nil)
        }
        assert_equal("foo=1",
            klass.new(@c1, @c1a).to_sql)
    end
    def test_and
        andor_paramtest(Compare::And)
        assert_equal("(foo=1 AND foo=2)",
            Compare::And.new(Compare::And.new(@c1, @c1a), @c2).to_sql)
        assert_equal("(foo=2 AND foo=1)",
            Compare::And.new(@c2, Compare::And.new(@c1, @c1a)).to_sql)
        comparator = 
        assert_equal("(foo=2 AND foo=1)",
            Compare::And.new(Compare::And.new(@c2, @c2a),
                             Compare::And.new(@c1, @c1a)).to_sql)
    end
    def test_or
        andor_paramtest(Compare::Or)
    end
end
